import sqlite3
import os
import pandas as pd

from . import household
from . import border_vicinity
from . import license_plate
from . import population
from . import income
from . import crime
from . import regions
from . import utils
from . import education
from . import election

FEATURE_LIST = [
    "license_plate", "income", "regions", "border_vicinity", "education", "election", "household", "population", "crime"]


class DataBase:
    """Interface to sqlite database stored in data/sqlite"""

    def __init__(self):
        """Initializing database, connecting to db, ..."""
        path = os.path.join(utils.path_to_data_dir(), "sqlite", "database.db")
        self.con = sqlite3.connect(path)

    def populate_db(self):
        """Populate database with processed data"""

        def load_data(feature):
            df = globals()[feature].load_data()
            df.to_sql(feature, self.con, if_exists="replace")
            return df

        for feature in FEATURE_LIST:
            load_data(feature)

    def query(self, sql_query):
        """Execute a query.

        Args:
            sql_query (str): sql  query

        Returns:
            DataFrame: Result of query
        """
        df = pd.read_sql(sql_query, self.con, index_col="index")
        return df

    def get_data(self):
        """Load data from db. For details on columns, see data/processed/data_desc.csv

        Returns:
            DataFrame: Data on regions, income, crime and population
        """

        df_merged = pd.DataFrame(columns=["kreis_key"])

        for feature in FEATURE_LIST:
            df_feat = self.query(f"SELECT * FROM {feature}")
            df_merged = df_merged.merge(df_feat, on="kreis_key", how="outer")

        # No Boolean data type in sqlite

        df_merged.east = df_merged.east.astype(bool)
        df_merged = df_merged.dropna(subset=["kreis_name"])
        df_merged.to_csv(os.path.join(
            utils.path_to_data_dir(), "processed", "merged.csv"))

        return df_merged



import os
import pandas as pd


def fix_key(ser):
    """Transform int to 5-digit str, padded with zeros. Ex: 101 -> 00101

    Args:
        ser (Series): Series of ints 

    Returns:
        Series: Series of strings, 5 digits + padded with zeros
    """
    return ser.astype(int).astype(str).str.zfill(5)


def to_numeric(df, except_=["kreis_key"]):
    """Convert all columns to numeric, except the columns passed via except_

    Args:
        df (pd.DataFrame): Target dataframe.
        except_ (list): List of columns not to convert. Optional, defaults to ["kreis_key"]. 

    Returns:
        pd.DataFrame: Dataframe with converted columns
    """
    df = df.copy()

    for col in df.columns:
        if col not in except_:
            df[col] = pd.to_numeric(df[col], errors="coerce")

    return df


def path_to_data_dir():
    """Path to Data directory

    Returns:
        str: Absolute path to data directory
    """
    return os.path.join(os.path.dirname(__file__), "data_files")


def fix_goettingen(df, col, proportional=False):
    """In 2016, regions 3152 and 3156 merged to become Göttingen, 3159

    Args:
        df (DataFrame): [description]
        col (str): [description]

    Returns:
        [type]: [description]
    """
    df.kreis_key = pd.to_numeric(df.kreis_key)
    df = df.set_index("kreis_key")
    if proportional:  # relative features should not be added naively
        # Rough approximation using the bigger county
        df.loc[3159, col] = pd.to_numeric(df.loc[3156, col], errors="coerce")
    else:  # absolute features can be added
        df.loc[3159, col] = pd.to_numeric(
            df.loc[3152, col], errors="coerce") + pd.to_numeric(df.loc[3156, col], errors="coerce")
    df = df.reset_index()
    df.kreis_key = fix_key(df.kreis_key)
    return df



import pandas as pd
import os
from . import utils


def prep_data():
    path = os.path.join(utils.path_to_data_dir(),
                        "raw", "education", "AI-N-06.csv")
    df_raw = pd.read_csv(path, encoding="ISO-8859-1",
                         skiprows=4, delimiter=";")
    df_raw.columns = ["year", "kreis_key",
                      "kreis_name", "prop_abitur", "prop_no_haupt"]

    df_raw.prop_abitur = pd.to_numeric(
        df_raw.prop_abitur.str.replace(",", "."), errors="coerce")
    df_raw.prop_no_haupt = pd.to_numeric(
        df_raw.prop_no_haupt.str.replace(",", "."), errors="coerce")

    df_raw = df_raw.dropna(subset=["kreis_key"])

    df_raw = df_raw[df_raw.kreis_key != "DG"]
    df_raw.kreis_key.replace(
        {"02": "02000", "11": "11000"}, inplace=True)  # Berlin, Hamburg

    df_raw = df_raw[[len(i) == 5 for i in df_raw.kreis_key]]

    # Pivot Data
    df_piv = df_raw.pivot(
        index="kreis_key", columns="year", values=["prop_abitur", "prop_no_haupt"])

    df_piv.columns = df_piv.columns.droplevel(0)
    df_piv = df_piv.reset_index()
    df_piv.columns = ["kreis_key"] + [f"prop_abitur_{i}" for i in df_raw.year.unique()] + [
        f"prop_no_haupt_{i}" for i in df_raw.year.unique()]

    for year in range(2013, 2017):
        df_piv = utils.fix_goettingen(
            df_piv, f"prop_abitur_{year}", proportional=True)
        df_piv = utils.fix_goettingen(
            df_piv, f"prop_no_haupt_{year}", proportional=True)

    # No data for Landkreis Bamberg and Schweinfurt

    df_piv = df_piv.fillna(df_piv.mean())

    return df_piv


def load_data():
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                                  "education", "education.csv"), index_col=0)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



import pandas as pd
import os
from . import utils
from . import license_plate


def prep_data():
    """Preprocess data on regions

    Returns:
        DataFrame: data on regions
    """
    df_plate = license_plate.load_data(with_kreis_name=True)
    df_regions = df_plate.drop(columns=["license_plate"])
    df_regions = df_regions[["kreis_key", "kreis_name"]].drop_duplicates()

    df_regions["bl_key"] = df_regions.kreis_key.str.slice(0, 2).astype(int)

    df_regions["east"] = df_regions.bl_key >= 12

    return df_regions[["kreis_key", "kreis_name", "east"]]


def load_data():
    """Load data on regions

    Returns:
        DataFrame: data on regions
    """
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                                  "regions", "regions.csv"), index_col=0)
    df.east = df.east.astype(bool)
    df.kreis_key = utils.fix_key(df.kreis_key)

    return df



import pandas as pd
import os
from . import utils
from . import population
from datetime import datetime

crime_categories = {
    "crimes": ["Straftaten insgesamt"],
    "fraud": [
        "Betrug §§ 263, 263a, 264, 264a, 265, 265a, 265b StGB davon:", "Betrug §§ 263, 263a, 264, 264a, 265, 265a, 265b StGB",
        "Urkundenfälschung §§ 267-271, 273-279, 281 StGB", "Betrug §§ 263, 263a, 264, 264a, 265, 265a-e StGB"],
    "violence": ["Gefährliche und schwere Körperverletzung, Verstümmelung weiblicher Genitalien §§ 224, 226, 226a, 231 StGB",
                 "Vorsätzliche einfache Körperverletzung § 223 StGB",
                 "Gewaltkriminalität",
                 "Tätlicher Angriff auf Vollstreckungsbeamte und gleichstehende Personen §§ 114, 115 StGB",
                 "Vergewaltigung, sexuelle Nötigung und sexueller Übergriff im besonders schweren Fall einschl. mit Todesfolge §§ 177, 178 StGB",
                 "Mord, Totschlag und Tötung auf Verlangen"],
    "theft": ["Raub, räuberische Erpressung und räuberischer Angriff auf Kraftfahrer §§ 249-252, 255, 316a StGB", "Raub, räuberische Erpressung auf/gegen Geldinstitute, Postfilialen und -agenturen",
              "Raub, räuberische Erpressung auf/gegen sonstige Zahlstellen und Geschäfte",
              "Handtaschenraub",
              "Sonstige Raubüberfälle auf Straßen, Wegen oder Plätzen",
              "Raubüberfälle in Wohnungen",
              "Diebstahl ohne erschwerende Umstände §§ 242, 247, 248a-c StGB und zwar:",
              "Einfacher Ladendiebstahl",
              "Diebstahl unter erschwerenden Umständen §§ 243-244a StGB und zwar:",
              "Wohnungseinbruchdiebstahl §§ 244 Abs. 1 Nr. 3 und Abs. 4, 244a StGB",
              "Tageswohnungseinbruchdiebstahl §§ 244 Abs. 1 Nr. 3 und Abs. 4, 244a StGB",
              "Diebstahl insgesamt und zwar:",
              "Diebstahl insgesamt von Kraftwagen einschl. unbefugte Ingebrauchnahme",
              "Diebstahl insgesamt von Mopeds und Krafträdern einschl. unbefugte Ingebrauchnahme",
              "Diebstahl insgesamt von Fahrrädern einschl. unbefugte Ingebrauchnahme",
              "Diebstahl insgesamt an/aus Kraftfahrzeugen",
              "Taschendiebstahl insgesamt"
              ],
    "drug": ["Rauschgiftdelikte (soweit nicht bereits mit anderer Schlüsselzahl erfasst)"]
}


def crime_filter(df, categories, column="Straftat"):
    """Construct filter for crimes listed in [categories]

    Args:
        df ([type]): [description]
        column (str, optional): [description]. Defaults to "Straftat".

    Returns:
        [type]: [description]
    """
    filt = df[column] == categories[0]
    for cat in categories:
        filt = filt | (df[column] == cat)
    return filt


def year_to_path(year):
    """Compute path of data on crimes for the given year

    Args:
        year (int): year

    Returns:
        str: path to data file
    """
    data_path = os.path.join(utils.path_to_data_dir(), "raw", "crime")
    path = str(year)
    files = os.listdir(os.path.join(data_path, "bka", path))
    if len(files) > 0:
        return os.path.join(data_path, "bka", path, files[0])


def prep_data_2013():
    """Preprocess data on crimes in 2013

    Returns:
        DataFrame: data on crimes in 2013
    """
    df = pd.read_excel(year_to_path(2013), skiprows=6)[
        ["Unnamed: 1", "Unnamed: 2", "Fälle"]].dropna(subset=["Unnamed: 2"])
    df.rename(columns={
        "Unnamed: 1": "Straftat", "Unnamed: 2": "kreis_key", "Fälle": "crimes_2013"}, inplace=True)
    cats = df.Straftat.unique()

    df.kreis_key = utils.fix_key(df.kreis_key)

    df = df[["Straftat", "kreis_key", "crimes_2013"]]

    df_ges = pd.DataFrame()

    for cat in crime_categories:
        df_cat = df[crime_filter(df, crime_categories[cat])]
        df_cat = df_cat.groupby("kreis_key").sum().reset_index()
        df_cat = df_cat.rename(columns={"crimes_2013": f"{cat}_2013"})
        if not df_ges.empty:
            df_ges = df_ges.merge(df_cat, on="kreis_key", how="outer")
        else:
            df_ges = df_cat

        df_ges = utils.fix_goettingen(df_ges, f"{cat}_2013")

    df_ges = utils.fix_goettingen(df_ges, "crimes_2013")

    return df_ges, list(cats)


def prep_data_14_20(year):
    """Preprocess data on crimes in the specified year

    Args:
        year (int): year in the range 2014-2020

    Returns:
        DataFrame: data on crimes in the given year
    """
    crime_clm = f"crimes_{year}"

    df = pd.read_csv(year_to_path(year), encoding="ISO-8859-1",
                     delimiter=";", skiprows=1, thousands=",")
    cats = df.Straftat.unique()
    df.rename(columns={"Gemeindeschlüssel": "kreis_key", "Anzahl erfasste Faelle": crime_clm,
              "erfasste Fälle": crime_clm, "Gemeindeschluessel": "kreis_key", "erfasste Faelle": crime_clm}, inplace=True)
    df.kreis_key = utils.fix_key(df.kreis_key)

    df_ges = pd.DataFrame()

    for cat in crime_categories:
        df_cat = df[["kreis_key", "Straftat", crime_clm]
                    ][crime_filter(df, crime_categories[cat])]
        df_cat = df_cat.groupby("kreis_key").sum().reset_index()
        df_cat = df_cat.rename(columns={crime_clm: f"{cat}_{year}"})
        if not df_ges.empty:
            df_ges = df_ges.merge(df_cat, on="kreis_key")
        else:
            df_ges = df_cat
        if year <= 2016:
            df_ges = utils.fix_goettingen(df_ges, f"{cat}_{year}")

    return df_ges, list(cats)


def prep_data():
    """Preprocess crime data

    Returns:
        DataFrame: crime data in the years 2013-2020
    """
    df, cats = prep_data_2013()

    for i in range(2014, 2021):
        df2, cats2 = prep_data_14_20(i)
        df = df.merge(df2, on="kreis_key", how="outer")
        cats = cats + cats2
    cats_df = pd.DataFrame(pd.Series(cats).unique())
    cats_df.to_csv(os.path.join(utils.path_to_data_dir(),
                   "processed", "crime", "categories.csv"))

    # calculate crime rates

    df_population = population.load_data()

    df_crime_rates = df.merge(df_population, on="kreis_key")
    years = list(filter(lambda y: f"population_{y}" in df_crime_rates.columns and f"crimes_{y}" in df_crime_rates.columns, range(2000, datetime.today(
    ).year+2)))

    cols = ["kreis_key"]

    for cat in crime_categories:
        for year in years:
            df_crime_rates[f"{cat}_pp_{year}"] = df_crime_rates[f"{cat}_{year}"] / \
                df_crime_rates[f"population_{year}"]
        cols = cols + [f"{cat}_{year}" for year in years]
        cols = cols + [f"{cat}_pp_{year}" for year in years]

    df_crime_rates = df_crime_rates[cols]

    return df_crime_rates


def load_data():
    """Load crime data from csv

    Returns:
       DataFrame : data on crimes
    """
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                                  "crime", "crime.csv"), index_col=0)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



import pandas as pd
from . import data
import os


def prep_data():
    """Preprocess the raw data in data/raw.

    Returns:
        DataFrame: data indexed by license plate codes
    """

    features = ["income", "population", "license_plate", "regions",
                "education", "crime", "household",  "election", "border_vicinity"]

    def prep_feature(feature):
        print(feature)
        df = getattr(data, feature).prep_data()
        df.to_csv(os.path.join(data.utils.path_to_data_dir(),
                  "processed", feature, f"{feature}.csv"))
        return df

    for feature in features:
        prep_feature(feature)



import pandas as pd
import os
from . import utils
from .. import geo


def prep_data():
    """Preprocess population data. Reads data in data/raw/population, outputs processed data in data/processed/population"""
    path = os.path.join(utils.path_to_data_dir(),
                        "raw", "population", "12411-05-01-4.csv")
    path_density = os.path.join(
        utils.path_to_data_dir(), "raw", "population", "AI002-1.csv")
    df_raw = pd.read_csv(path, encoding="ISO-8859-1",
                         skiprows=6, delimiter=";")

    df_dens = pd.read_csv(path_density, encoding="ISO-8859-1",
                          delimiter=";", skiprows=4)

    # clean data

    df_raw.rename(columns={"Unnamed: 0": "year", "Unnamed: 1": "kreis_key",
                  "Unnamed: 2": "kreis_name", "Insgesamt": "population"}, inplace=True)
    df_raw = df_raw.dropna(subset=["kreis_key"])
    df_raw = df_raw[df_raw.kreis_key != "DG"]
    df_raw.kreis_key.replace(
        {"02": "02000", "11": "11000"}, inplace=True)  # Berlin, Hamburg

    df_dens = df_dens.iloc[:, :4]
    df_dens.columns = ["year", "kreis_key", "kreis_name", "population_density"]
    df_dens = df_dens[["year", "kreis_key", "population_density"]]
    df_dens = df_dens.dropna(subset=["kreis_key"])
    df_dens.kreis_key.replace(
        {"02": "02000", "11": "11000"}, inplace=True)  # Berlin, Hamburg
    df_dens.population_density = pd.to_numeric(
        df_dens.population_density.str.replace(",", "."), errors="coerce")

    df_raw = df_raw.merge(df_dens, on=["kreis_key", "year"])

    # Pivot Data
    df_piv = df_raw.pivot(
        index="kreis_key", columns="year", values=["population", "population_density"])

    df_piv.columns = [f"{desc}_{year}" for desc, year in df_piv.columns]

    df_piv = df_piv.reset_index()

    df = df_piv[[len(i) == 5 for i in df_piv.kreis_key]].copy()

    for year in range(2013, 2016):
        df = utils.fix_goettingen(df, f"population_{year}")
        df = utils.fix_goettingen(
            df, f"population_density_{year}", proportional=True)

    df = df.dropna(subset=["kreis_key"])

    return df


def load_data():
    """Load data from csv stored in data/processed"""
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                                  "population", "population.csv"), index_col=0)
    df = utils.to_numeric(df)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



try:
    from . import visualization

    app = visualization.app.VisApp()
    server = app.app.server
except:
    pass

from .data import db


def load_regional_data():
    """ Load data frame"""
    return db.get_data()



import pandas as pd
from bs4 import BeautifulSoup
import requests
import re
import os
from . import utils


def prep_data():
    wiki_kreise = requests.get(
        "https://de.wikipedia.org/wiki/Liste_der_Landkreise_in_Deutschland")

    soup = BeautifulSoup(wiki_kreise.text, features="html.parser")

    df_kreise = pd.DataFrame()

    for tr in soup.find_all('tr'):
        """Format: <tr>
                    <td>REGIONAL_SCHLÜSSEL</td>
                    <td>LANDKREIS</td>
                    <td>KFZ-KENNZ</td>
                    <td>BUNDESLAND</td>
                    <td>KREISSITZ</td>
                    <td>EINWOHNER</td>
                    <td>FLÄCHE</td>
                    <td>BEVÖLKERUNGSDICHTE</td>
                    <td>KARTE</td></tr>
                    <tr>"""
        tds = tr.find_all('td')
        if(len(tds) > 5):
            kreis_key = tds[0].text.strip()
            kreis_name = tds[1].text.split(",")[0].strip()
            license_plates = tds[2].text.replace(",", " ").replace(
                "(", " ").replace(")", " ").split()

            for license_plate in license_plates:
                df_kreise = df_kreise.append(
                    {"license_plate": license_plate.strip(), "kreis_key": kreis_key, "kreis_name": kreis_name}, ignore_index=True)

    wiki_städte = requests.get(
        "https://de.wikipedia.org/wiki/Liste_der_kreisfreien_St%C3%A4dte_in_Deutschland")

    soup = BeautifulSoup(wiki_städte.text, features="html.parser")

    df_städte = pd.DataFrame()

    for tr in soup.find_all('tr'):
        """ Format:
        <tr>
            <td>WAPPEN</td>
            <td>STADT</td>
            <td>REGIONALSCHLÜSSEL
            </td>
            <td>BUNDESLAND</td>
            <td>RGIERUNGSBEZIRK</td>
            <td>KFZ-KENNZEICHEN
            </td>
            <td>FLÄCHE
            </td>
            <td>TEINWOHNER 1939
            </td>
            <td>TEINWOHNER 1950
            </td>
            <td>TEINWOHNER 1970
            </td>
            <td>TEINWOHNER 1990
            </td>
            <td>TEINWOHNER 2011
            </td>
            <td>EINWOHNER 2020</small>
            </td>
            <td>BEVDICHTE
            </td>
            <td>KARTE</td></tr>
        """
        tds = tr.find_all('td')
        if(len(tds) > 12):
            kreis_key = tds[2].text.strip()
            kreis_name = re.split("\(|\[", tds[1].text)[0].strip()
            license_plates = tds[5].text.replace(",", " ").replace(
                "(", " ").replace(")", " ").split()

            if(all(df_kreise["kreis_key"] != kreis_key)):  # Avoid duplicates
                for license_plate in license_plates:
                    df_städte = df_städte.append(
                        {"license_plate": license_plate.strip(), "kreis_key": kreis_key, "kreis_name": kreis_name}, ignore_index=True)

    df = pd.concat([df_städte, df_kreise])
    # Fix regional code of Berlin ("11000")
    df["kreis_key"].replace({"11001": "11000"}, inplace=True)

    df.kreis_key = utils.fix_key(df.kreis_key)

    df = df.append({"license_plate": "EA", "kreis_key": "16056",
                    "kreis_name": "Eisenach"}, ignore_index=True)  # Manually add Eisenach

    return df


def load_data(with_kreis_name=False):
    """Load data from csv stored in data/processed"""
    cols = ["kreis_key", "license_plate"]
    if with_kreis_name:
        cols.append("kreis_name")
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                     "license_plate", "license_plate.csv"), index_col=0, dtype={"kreis_key": "object"})[cols]
    return df



from . import regions
from . import utils
from .. import geo
import os
import pandas as pd


def prep_data():
    """Preprocess data on border vicinity
    """
    print("This may take a while!")
    df = regions.load_data()
    df = df[["kreis_key"]]
    df.drop_duplicates()
    distance_calculator = geo.border_vicinity.DistanceCalculator()
    df["border_vic"] = df.kreis_key.apply(distance_calculator.distance)
    return df[["kreis_key", "border_vic"]]


def load_data():
    """Load data on border vicinity
    """
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                     "border_vicinity", "border_vicinity.csv"), index_col=0)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



import pandas as pd
import os
from . import utils


def prep_data():
    """Preprocess data on income

    Returns:
        DataFrame: data on income
    """
    path = os.path.join(utils.path_to_data_dir(),
                        "raw", "income", "82411-01-03-4.csv")
    df_raw = pd.read_csv(path, encoding="ISO-8859-1",
                         skiprows=6, delimiter=";")

    df_raw = df_raw.rename(columns={"Unnamed: 0": "year", "Unnamed: 1": "kreis_key",
                           "Unnamed: 2": "kreis_name", "Tsd. EUR": "income", "EUR": "income_pp"})
    df_raw = df_raw.dropna(subset=["kreis_key"])
    df_raw = df_raw[df_raw.kreis_key != "DG"]
    df_raw.kreis_key.replace(
        {"02": "02000", "11": "11000"}, inplace=True)  # Berlin, Hamburg

    df_piv = df_raw.pivot(index="kreis_key", columns="year",
                          values=["income_pp", "income"])

    df_piv.columns = [f"{s1}_{s2}" for (s1, s2) in df_piv.columns.tolist()]

    df_piv.reset_index(inplace=True)

    # drop rows that are not on "Kreis"-level
    df = df_piv[[len(i) == 5 for i in df_piv.kreis_key]]

    return df


def load_data():
    """Load data from csv stored in data/processed"""
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(), "processed",
                     "income", "income.csv"), index_col=0)

    df = utils.to_numeric(df)

    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



feature_info = {
    "crimes_pp": {
        "title": "Erfasste Straftaten",
        "label": "Straftaten / EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "fraud_pp": {
        "title": "Betrugsfälle",
        "label": "Betrugsfälle / EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "theft_pp": {
        "title": "Diebstähle",
        "label": "Diebstähle / EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "violence_pp": {
        "title": "Gewaltkriminalität",
        "label": "Straftaten / EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "drug_pp": {
        "title": "Rauschgiftdelikte",
        "label": "Fälle / EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "income_pp": {
        "title": "Verfügbares Einkommen der privaten Haushalte",
        "label": "Euro / EW",
        "time_dep": True,
        "color": "gray"
    },
    "hh_avg": {
        "title": "Durchschnittliche Haushaltsgröße (2011)",
        "label": "Personen",
        "time_dep": False,
        "color": "gray_r"
    },
    "population": {
        "title": "Bevölkerung",
        "label": "EW",
        "time_dep": True,
        "color": "gray_r"
    },
    "population_density": {
        "title": "Bevölkerungsdichte",
        "label": "EW / qkm",
        "time_dep": True,
        "color": "gray_r"
    },
    "prop_abitur": {
        "title": "Anteil Schulabgänger/-innen mit allgem. Hochschulreife",
        "label": "%",
        "time_dep": True,
        "color": "gray_r"
    },
    "prop_no_haupt": {
        "title": "Anteil Schulabgänger/-innen ohne Hauptschulabschluss",
        "label": "%",
        "time_dep": True,
        "color": "gray_r"
    },
    "border_vic": {
        "title": "Entfernung zur Grenze",
        "label": "Km",
        "time_dep": False,
        "color": "gray_r"
    },
    "ew_vot_rel_alternative_fuer_deutschland_2019": {
        "title": "AFD-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": "blues"
    },
    "ew_vot_rel_sozialdemokratische_partei_deutschlands_2019": {
        "title": "SPD-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": "reds"
    },
    "ew_vot_rel_buendnis_90_die_gruenen_2019": {
        "title": "Grüne-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": "greens"
    },
    "ew_vot_rel_christlich_demokratische_union_deutschlands_2019": {
        "title": "CDU/CSU-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": "greys"
    },
    "ew_vot_rel_freie_demokratische_partei_2019": {
        "title": "FDP-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": ["white", "yellow"]
    },
    "ew_vot_rel_die_linke_2019": {
        "title": "Linke-Stimmanteil, Europawahl 2019",
        "label": "%",
        "time_dep": False,
        "color": "reds"
    },
    "ew_winner_2019": {
        "title": "Europa-Wahl 2019",
        "label": "Partei",
        "time_dep": False,
        "color": {
            "alternative_fuer_deutschland": "lightblue",
            'sozialdemokratische_partei_deutschlands': "red",
            'buendnis_90_die_gruenen': "lightgreen",
            'christlich_demokratische_union_deutschlands': "darkgray",
            "freie_demokratische_partei": "yellow",
            "die_linke": "red"
        }
    }
}



import pandas as pd
import os
from . import utils


def prep_data():
    path = os.path.join(utils.path_to_data_dir(),
                        "raw", "household", "12111-31-01-4.csv")
    df_raw = pd.read_csv(path, encoding="ISO-8859-1",
                         skiprows=9, delimiter=";")

    df_raw.columns = ["kreis_key", "kreis_name", "hh_ges",
                      "hh_1", "hh_2", "hh_3", "hh_4", "hh_5", "hh_6"]

    df_raw = df_raw[df_raw.kreis_key != "DG"]

    # Change in regions since 2011:

    region_replace = {
        "13001": "13075",
        "13002": "13071",
        "13005": "13073",
        "13006": "13074",
        "13051": "13072",
        "13052": "13071",
        "13053": "13072",
        "13054": "13076",
        "13055": "13071",
        "13057": "13073",
        "13058": "13074",
        "13059": "13075",
        "13060": "13076",
        "13061": "13073",
        "13062": "13075",
        "03152": "03159",  # GÖTTTINGEN
        "03156": "03159"
    }

    df_raw.kreis_key.replace(region_replace, inplace=True)

    df_raw.kreis_key.replace(
        {"02": "02000", "11": "11000"}, inplace=True)  # Berlin, Hamburg

    df_raw = df_raw[[len(i) == 5 for i in df_raw.kreis_key]]

    df_raw = utils.to_numeric(df_raw)

    df_raw = df_raw.groupby("kreis_key").sum().reset_index()
    df_raw = df_raw.set_index("kreis_key")
    df_raw = df_raw.reset_index()

    df_raw["hh_avg"] = (df_raw["hh_1"] + 2*df_raw["hh_2"] + 3*df_raw["hh_3"] + 4 *
                        df_raw["hh_4"] + 5*df_raw["hh_5"] + 6*df_raw["hh_6"]) / df_raw["hh_ges"]

    return df_raw[["kreis_key", "hh_ges", "hh_avg"]]


def load_data():
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(),
                     "processed", "household", "household.csv"), index_col=0)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df



import pandas as pd
import os
from . import utils


def prep_data():
    """Preprocess data on the Europawahl 2019 / 2014
    """
    election_path = os.path.join(
        utils.path_to_data_dir(), "raw", "election", "ew19_kerg.csv")
    election = pd.read_csv(election_path, skiprows=[0, 1, 3, 4], delimiter=";")
    election.dropna(axis=1, inplace=True, how="all")

    general_columns = ["kreis_key",  "kreis_name", "bl_key", "ew_eli_2019", "ew_eli_2014", "ew_vot_2019", "ew_vot_2014",
                       "ew_invalid_2019", "ew_invalid_2014", "ew_valid_2019", "ew_valid_2014"]

    election.columns = general_columns + list(election.columns[11:])
    election.drop("kreis_name", axis=1, inplace=True)
    election.dropna(subset=["kreis_key"], inplace=True)

    general_columns.remove("kreis_name")

    df_partei = pd.DataFrame()

    new_columns = list(election.columns[:10])
    partei_key = ""
    partei_keys = []
    for i in range(10, len(election.columns)):
        clm = election.columns[i]
        if "Unnamed: " not in clm:
            partei_key = clm.lower()
            replacements = {
                " ": "_",
                "/": "_",
                ".": "_",
                "-": "_",
                "ä": "ae",
                "ö": "oe",
                "ü": "ue",
                "ß": "ss"
            }
            for rep in replacements:
                partei_key = partei_key.replace(rep, replacements[rep])

            df_partei = df_partei.append(
                {"partei_key": partei_key, "partei_name": clm}, ignore_index=True)
            partei_keys.append("")

            new_columns.append(f"ew_vot_abs_{partei_key}_2019")
        else:
            new_columns.append(f"ew_vot_abs_{partei_key}_2014")

    df_partei.to_csv(os.path.join(utils.path_to_data_dir(),
                     "processed", "election", "parties.csv"))

    election.columns = new_columns

    election = election[election.kreis_key > 99]

    election.kreis_key = utils.fix_key(election.kreis_key)

    election = election.fillna(0)

    interesting_parties = ["alternative_fuer_deutschland",
                           'sozialdemokratische_partei_deutschlands', 'buendnis_90_die_gruenen', 'christlich_demokratische_union_deutschlands', "freie_demokratische_partei", "die_linke"]

    # combine CDU / CSU

    for y in [2014, 2019]:
        election[f'ew_vot_abs_christlich_demokratische_union_deutschlands_{y}'] = election[f'ew_vot_abs_christlich_demokratische_union_deutschlands_{y}'] + \
            election[f'ew_vot_abs_christlich_soziale_union_in_bayern_e_v__{y}']
        for party in interesting_parties:
            election[f"ew_vot_rel_{party}_{y}"] = election[f'ew_vot_abs_{party}_{y}'] / \
                election[f"ew_vot_{y}"]
        party_vot_cols = [
            f"ew_vot_rel_{party}_{y}" for party in interesting_parties]

        election[f"ew_winner_{y}"] = election[party_vot_cols].idxmax(
            axis=1).str.slice(11, -5)

    cols_to_return = general_columns + \
        [f"ew_vot_{relabs}_{party}_{y}" for party in interesting_parties for relabs in [
            "rel", "abs"] for y in [2014, 2019]] + ["ew_winner_2014"] + ["ew_winner_2019"]

    return election[cols_to_return]


def load_data():
    df = pd.read_csv(os.path.join(utils.path_to_data_dir(),
                     "processed", "election", "election.csv"), index_col=0)
    df.kreis_key = utils.fix_key(df.kreis_key)
    return df




from re import S
import dash
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objects as go
import plotly.express as px
import pandas as pd
import geopandas as gpd
import dash_bootstrap_components as dbc
from .. import geo
from .. import data
from . import config


class VisApp:
    """Dash application to visualize collected data"""

    def __init__(self):
        """Setup Dash aplication
        """
        # setup data
        self.df = self.data_for_map()

        # configuration for visualization

        self.columns = config.feature_info

        # creating a dash app
        self.app = dash.Dash(__name__, external_stylesheets=[
                             dbc.themes.BOOTSTRAP, "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.5.0/font/bootstrap-icons.css"])
        self.app.title = "IF_LICENSE_PLATES_COULD_TALK"
        self.app.layout = dbc.Container([
            dcc.Location(id="url", refresh=False),
            html.H1(children="IF_LICENSE_PLATES_COULD_TALK",
                    style={"margin-top": "30px"}),
            dbc.Tabs(id="page-content", children=[
                dbc.Tab(label="Maps", children=self.generate_map_page()),
                dbc.Tab(label="Scatter plots",
                        children=self.generate_scatter_page()),
                dbc.Tab(label="Time series",
                        children=self.generate_time_series_page())
            ]),
        ])
        self.setup_scatter_callbacks()
        self.setup_map_callbacks()
        self.setup_timeseries_callback()

    def run(self):
        """Start Dash server in debug mode

        Returns:
            None
        """
        self.app.run_server()

    def generate_map_page(self):
        return dbc.Container([
            dbc.Container([
                html.P("Feature:"),
                dcc.Dropdown(
                    id="map_feature_select",
                    options=[{"label": self.columns[feature]["title"],
                              "value": feature} for feature in self.columns],
                    value="crimes_pp"
                )],  style={"margin-top": "20px"}),
            dbc.Container([
                html.P("Year:"),
                dcc.Slider(id="map_year", min=2013, max=2018, value=2018, marks={y: f"{y}" for y in range(2013, 2019)})], style={"margin-top": "20px"}),
            html.Hr(),
            dcc.Loading(id="map_loading", type="circle",
                        children=[dbc.Container(id="map_output")])])

    def generate_scatter_page(self):

        return dbc.Container(
            [dbc.Row([dbc.Col(children=[
                html.P("Feature (x-axis):"),
                dcc.Dropdown(
                    id="scatter_feature_x",
                    options=[{"label": self.columns[feature]["title"],
                              "value": feature} for feature in self.columns],
                    value="crimes_pp"
                )]), dbc.Col(children=[html.P("Scale (x-axis):"),
                                       dcc.Dropdown(
                    id="scatter_log_x", options=[{"label":  "linear",
                                                  "value": "linear"}, {"label": "log", "value": "log"}],
                    value="linear"
                )])], style={"margin-top": "20px"}),
                dbc.Row([dbc.Col(children=[
                    html.P("Feature (y-axis):"),
                    dcc.Dropdown(
                        id="scatter_feature_y",
                        options=[{"label": self.columns[feature]["title"],
                                  "value": feature} for feature in self.columns],
                        value="income_pp"
                    )]), dbc.Col(children=[html.P("Scale (y-axis):"),
                                           dcc.Dropdown(
                        id="scatter_log_y", options=[{"label":  "linear",
                                                      "value": "linear"}, {"label": "log", "value": "log"}],
                        value="linear"
                    )])],  style={"margin-top": "20px"}),
             dbc.Container([
                 html.P("Year:"),
                 dcc.Slider(id="scatter_year", min=2013, max=2018, value=2018, marks={y: f"{y}" for y in range(2013, 2019)})], style={"margin-top": "20px"}),
             html.Hr(),
             dcc.Loading(id="scatter_loading", type="circle",
                         children=[dbc.Container(id="scatter_output")])
             ])

    def generate_scatter_plot(self, feature_x, log_x,  feature_y, log_y, year):
        featinfo_x = self.columns[feature_x]
        col_x = feature_x
        if featinfo_x["time_dep"]:
            col_x += f"_{year}"
        log_x = log_x == "log"

        featinfo_y = self.columns[feature_y]
        col_y = feature_y
        if featinfo_y["time_dep"]:
            col_y += f"_{year}"
        log_y = log_y == "log"

        fig = px.scatter(data_frame=self.df,
                         x=col_x, y=col_y, color="east",  hover_data=["kreis_name"],
                         labels={col_x: featinfo_x["label"],
                                 col_y: featinfo_y["label"]},
                         log_x=log_x,
                         log_y=log_y,
                         trendline="ols",
                         trendline_options=dict(log_x=log_x, log_y=log_y),
                         trendline_color_override="red")

        return dcc.Graph(
            id='scatter',
            figure=fig
        )

    def generate_timeseries_plot(self, feature, kreis):
        featinfo = self.columns[feature]

        df = self.df[self.df.kreis_key == kreis].melt()
        df.columns = ["year", feature]
        df = df[df.year.str.contains(feature+"_2")].copy()
        df.year = pd.to_numeric(df.year.str.slice(start=-4))

        df.dropna()

        fig = px.line(data_frame=df, x="year", y=feature,
                      labels={feature: featinfo["label"]})

        return dcc.Graph(
            id="timeseries",
            figure=fig
        )

    def generate_time_series_page(self):
        return dbc.Container([
            dbc.Container([
                html.P("Feature:"),
                dcc.Dropdown(
                    id="timeseries_feature",
                    options=[{"label": self.columns[feature]["title"],
                              "value": feature} for feature in self.columns if self.columns[feature]["time_dep"]],
                    value="crimes_pp"
                )],  style={"margin-top": "20px"}),
            dbc.Container([
                html.P("Region:"),
                dcc.Dropdown(
                    id="timeseries_kreis",
                    options=[{"label": row[1],
                              "value": row[0]} for row in self.df.sort_values(by="kreis_name")[["kreis_key", "kreis_name"]].values],
                    value="02000"
                )],  style={"margin-top": "20px"}),
            html.Hr(),
            dcc.Loading(id="timeseries_loading", type="circle",
                        children=[dbc.Container(id="timeseries_output")])])

    def generate_map_output(self, feature, year):
        """Generate output structure

        Args:
            feature (str): feature column to display
            year (int): -

        Returns:
            [Component]: dash components to display map of germany with selected regional data.
        """
        return [
            dbc.Row([
                dbc.Col([
                    html.Div(id="map_container",
                             children=[self.generate_map(feature, year)])
                ])
            ])
        ]

    def setup_map_callbacks(self):
        """Setup the callbacks for the map page

        Returns:
            None
        """
        @ self.app.callback(
            dash.dependencies.Output("map_output", "children"),
            [dash.dependencies.Input("map_feature_select", "value"), dash.dependencies.Input("map_year", "value")])
        def update(feature, year):
            if feature != None:
                return self.generate_map_output(feature, year)

        @ self.app.callback(
            dash.dependencies.Output("map_year", "disabled"),
            [dash.dependencies.Input("map_feature_select", "value")])
        def update(feature):
            if feature != None:
                return not self.columns[feature]["time_dep"]

    def setup_scatter_callbacks(self):
        """Setup the callbacks for the scatter page
        """
        @ self.app.callback(
            dash.dependencies.Output("scatter_output", "children"),
            [dash.dependencies.Input("scatter_feature_x", "value"), dash.dependencies.Input("scatter_log_x", "value"), dash.dependencies.Input(
                "scatter_feature_y", "value"), dash.dependencies.Input("scatter_log_y", "value"), dash.dependencies.Input("scatter_year", "value")])
        def update(feature_x, log_x, feature_y, log_y, year):
            if None not in [feature_x, log_x, feature_y, log_y]:
                return self.generate_scatter_plot(feature_x, log_x,  feature_y, log_y,  year)

    def setup_timeseries_callback(self):
        """Setup the callbacks for the scatter page
        """
        @ self.app.callback(
            dash.dependencies.Output("timeseries_output", "children"),
            [dash.dependencies.Input("timeseries_feature", "value"), dash.dependencies.Input(
                "timeseries_kreis", "value")])
        def update(feature, kreis):
            if None not in [feature, kreis]:
                return self.generate_timeseries_plot(feature, kreis)

    def data_for_map(self):
        """Load data for visualization"""
        df_geo = geo.utils.load_geodata()
        df = data.db.get_data()

        df_comb = df_geo.merge(
            df, left_on="RS", right_on="kreis_key", how="left")

        return df_comb

    def generate_map(self, feature, year):
        """Generate map for visualization

        Args:
            feature (str): feature column
            year (int): year

        Returns:
            dcc.Graph: Map component
        """

        feature_info = self.columns[feature]

        if feature_info["time_dep"]:
            col = f"{feature}_{year}"
        else:
            col = feature

        fig = px.choropleth(self.df.fillna(0), geojson=self.df.geometry, locations=self.df.index, color=col, scope="europe",
                            color_continuous_scale=feature_info["color"],
                            range_color=(self.df[col].min(
                            ), self.df[col].max()),
                            hover_name="kreis_name",
                            hover_data=[col],
                            color_discrete_map=feature_info["color"],
                            labels={
                                col: feature_info["label"]
        })
        fig.update_geos(fitbounds="locations", visible=False)
        fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0})
        fig.update_layout(hoverlabel={"bgcolor": "white"})

        return dcc.Graph(
            id='map',
            figure=fig
        )



from . import prep_data
from . import setup_db
import argparse

parser = argparse.ArgumentParser(description="IF_LICENSE_PLATES_COULD_TALK")
parser.add_argument("cmd", nargs="?", default="run")

args = parser.parse_args()

if args.cmd == "prep":
    prep_data.prep_data()
elif args.cmd == "db":
    setup_db.setup_db()
else:
    from . import app
    app.run()



import geopy.distance
from shapely.geometry import LinearRing, shape
from . import utils
from .. import data
import geopandas as gpd
import pandas as pd
import numpy as np
import plotly.express as px


class DistanceCalculator():
    """Class to calculate distance between county and border of germany
    """

    def __init__(self):
        """Initialize
        """
        # load geodata
        self.df_geo = utils.load_geodata()
        # this projection preserves distance better than the one we use for visualization
        self.crs = "epsg:32662"
        # create border of germany
        germany = self.df_geo.dissolve()
        buffered = germany.to_crs(self.crs).buffer(10000).to_crs(germany.crs)
        self.border = buffered.boundary.to_crs(self.crs).explode().iloc[0]
        self.border_shape = shape(self.border)
        self.pol_ext = LinearRing(self.border_shape.coords)

    def distance(self, kreis_key):
        """Calculate distance between county and border"""
        # calculate midpoint of county
        kreis_geom = self.df_geo[self.df_geo["RS"] == kreis_key]
        center_point = kreis_geom.to_crs(self.crs).centroid
        center_point.reset_index(drop=True, inplace=True)

        # Calculate nearest point on boundary
        # https://stackoverflow.com/questions/33311616/find-coordinate-of-the-closest-point-on-polygon-in-shapely/33324058#33324058
        d = self.pol_ext.project(center_point.geometry.loc[0])
        p = self.pol_ext.interpolate(d)
        cp_x, cp_y = center_point.geometry.loc[0].coords[0]
        closest = list(p.coords)[0]

        # convert to original crs
        point_geometry = gpd.points_from_xy(
            x=[cp_x, closest[0]], y=[cp_y, closest[1]])
        point_df = gpd.GeoDataFrame(
            {}, geometry=point_geometry, crs=self.crs).to_crs(self.df_geo.crs)

        # use geopy to return geodesic distance
        return geopy.distance.geodesic(point_df.loc[0, "geometry"].coords[0], point_df.loc[1, "geometry"].coords[0]).kilometers



from . import regions
from . import population
from . import license_plate
from . import income
from . import household
from . import database
from . import utils
from . import border_vicinity
from . import crime
from . import education
from . import election

db = database.DataBase()



import geopandas as gpd
import os
import pyproj


def load_geodata():
    """Load geodata from shapefile

    Returns:
        gdp.GeoDataFrame:  GeoDataFrame containing regional data
    """
    df_geo = gpd.read_file(os.path.join(os.path.dirname(
        __file__),  "shapefiles", "VG250_KRS3", "VG250_KRS.shp"))
    df_geo.to_crs(pyproj.CRS.from_epsg(4326), inplace=True)
    return df_geo
